import argparse

# Flags
from .flags.todol_path import TodolPath
from .flags.todol_list import TodolList
from .flags.todol_upgrade import TodolUpgrade
from .flags.todol_version import TodolVersion

# Functions for the main loop
from .functionality.functions import Functions
from .functionality.prompts import Prompts
from .functionality.commands_list import COMMANDS, ALIASES
from .functionality.commands import Commands
from .functionality.paths import reset_todolist, backup_todolist, load_backup

def parse_args():
    parser = argparse.ArgumentParser(
        prog="todol",
        description=f"{TodolVersion.version()}\nSimple cli todo app",
        formatter_class=argparse.RawTextHelpFormatter
    )

    actions = parser.add_argument_group("Task actions")
    actions.add_argument("-a", "--add", nargs="+", metavar="TASK", help="Add new task")
    actions.add_argument("-d", "--done", nargs="+", metavar="ID", help="Mark task as done")
    actions.add_argument("-c", "--clear", action="store_true", help="Remove completed tasks")
    actions.add_argument("-o", "--order", action="store_true", help="Order Todos by ids")

    info = parser.add_argument_group("Information")
    info.add_argument("-ls", "--list", action="store_true", help="List tasks to the terminal")
    info.add_argument("-p", "--path", action="store_true", help="Show todo files in local directory")
    info.add_argument("-u", "--update", action="store_true", help="Update todol with pipx")
    info.add_argument("-up", "--upgrades", action="store_true", help="There's only one way to find out :)")
    info.add_argument("-v", "--version", action="store_true", help="Show version")

    file_action = parser.add_argument_group("File actions")
    file_action.add_argument("-rst", "--reset", action="store_true", help="Reset Todo list")
    file_action.add_argument("-bk", "--backup", action="store_true", help="Create backup")
    file_action.add_argument("-lbk", "--load", action="store_true", help="Load backup")

    return parser.parse_args()

def main():
    args = parse_args()

    # file actions

    if args.reset:
        reset_todolist()
        return

    if args.backup:
        backup_todolist()
        return

    if args.load:
        load_backup()
        return

    # commands
    if args.upgrades:
        from .flags.todol_upgrades import main
        main()
        return

    if args.add:
        Commands.cmd_add(args.add)
        return

    if args.done:
        Commands.cmd_done(args.done)
        return

    if args.clear:
        Functions.clearTaskJson()
        return

    if args.order:
        Functions.orderList()
        return

    # Flag flags

    if args.path:
        TodolPath.path()
        return

    if args.list:
        TodolList.list()
        return

    if args.update:
        TodolUpgrade.upgrade()
        return

    if args.version:
        print(TodolVersion.version())
        return

    # main loop

    Functions.greetingAppStart()

    while True:
        try:
            raw = Prompts.session.prompt('todol ~ $ ').strip()
        except KeyboardInterrupt:
            break

        if not raw:
            continue

        parts = raw.split()
        command, *args = parts
        command = command.lower()

        command = ALIASES.get(command, command)

        func = COMMANDS.get(command)

        if not func:
            print(f'{command}: command not found')
            continue

        try:
            func(args)
        except IndexError:
            print('Missing argument')
        except (SystemExit, KeyboardInterrupt):
            break




from .functions import Functions
from .prompts import Prompts

from rich import print

from prompt_toolkit.formatted_text import HTML

class Commands():
    def cmd_add(args):
            args = args or []

            try:
                task = " ".join(args)
            except IndexError:
                print("Add something bruhh")

            Functions.build_task(task)

    def cmd_done(args):
        Functions.doneTaskJson(args)

    def cmd_edit(args):
        try:
            taskId = args[0]
            task = Functions.getAllTasks()

            defTask: str = task[taskId]['task']

            editTask = Prompts.session.prompt(HTML(f'\n<ansiblue>todol ~ {taskId}: </ansiblue>'), default=defTask)

            Functions.update_task(taskId, editTask)

            print(f'\n[bold yellow]Task {taskId} Edited![/bold yellow]\n')

        except ValueError:
            print('Invalid input. Please enter a valid number.')
        except KeyError:
            print('Invalid input. Please enter a valid number.')

    def cmd_help(args):
        Functions.helpText()

    def cmd_list(args):
        Functions.openJson()

    def cmd_clear(args):
        Functions.clearTaskJson()

    def cmd_reload(args):
        Functions.greetingAppStart()

    def cmd_order(args):
        Functions.orderList()

    def cmd_exit(args):
        raise SystemExit



import sys
import requests
from importlib.metadata import version, PackageNotFoundError

PACKAGE_NAME = "todol"


class TodolVersion():
    def version():
        try:
            installed_version = version(PACKAGE_NAME)
        except PackageNotFoundError:
            installed_version = "not installed"

        try:
            resp = requests.get(f"https://pypi.org/pypi/{PACKAGE_NAME}/json", timeout=3)
            resp.raise_for_status()
            latest_version = resp.json()["info"]["version"]
        except Exception:
            latest_version = "unknown"

        if installed_version == latest_version:
            return (f"{PACKAGE_NAME} {installed_version} LTS")
        else:
            return (f"{PACKAGE_NAME} {installed_version} newer version available: {latest_version}")

        sys.exit(0)



from .paths import todoJsonListPath

import json
import re
from prompt_toolkit.shortcuts import clear
from rich.text import Text

from rich.console import Console
from rich import print

from collections import defaultdict

class Functions():

    # greeting
    # reload view

    def greetingAppStart():

        clear()

        print(r"""
   ▄▄▄█████▓ ▒█████   ▓█████▄  ▒█████   ██▓
   ▓  ██▒ ▓▒▒██▒  ██▒ ▒██▀ ██▌▒██▒  ██▒▓██▒
   ▒ ▓██░ ▒░▒██░  ██▒ ░██   █▌▒██░  ██▒▒██░
   ░ ▓██▓ ░ ▒██   ██░ ░▓█▄   ▌▒██   ██░▒██░
     ▒██▒ ░ ░ ████▓▒░ ░▒████▓ ░ ████▓▒░░██████▒
     ▒ ░░   ░ ▒░▒░▒░   ▒▒▓  ▒ ░ ▒░▒░▒░ ░ ▒░▓  ░
       ░      ░ ▒ ▒░   ░ ▒  ▒   ░ ▒ ▒░ ░ ░ ▒  ░
     ░      ░ ░ ░ ▒    ░ ░  ░ ░ ░ ░ ▒    ░ ░
                ░ ░      ░        ░ ░      ░  ░
"""
)
        print("[dim]      Type [bold yellow]help[/bold yellow] to see available commands[/dim]\n")
        Functions.openJson()

    def getAllTasks() -> dict:
        data = Functions.load_todos()
        return data['tasks']

    def update_task(task_id: str, task: str):

            data = Functions.load_todos()
            data['tasks'][task_id] = {
                "task": task,
                "completed": False,
            }
            Functions.save_todos(data)

    # open Json (write on start)

    def openJson() -> None:
        TAG_RE = re.compile(r'@(\w+)')
        console = Console()
        tasks = Functions.getAllTasks()

        grouped = defaultdict(list)

        for task_id, task in tasks.items():
            raw_text = task.get("task", "")
            completed = task.get("completed", False)

            tags = TAG_RE.findall(raw_text) or ["untagged"]
            clean_text = TAG_RE.sub("", raw_text).strip()

            for tag in tags:
                grouped[tag].append({
                    "id": task_id,
                    "text": clean_text,
                    "completed": completed,
                })

        for tag, items in grouped.items():
            items.sort(key=lambda t: t["completed"])
            completed_count = sum(1 for t in items if t["completed"])
            incomplete_count = len(items) - completed_count

            console.print()
            console.print(Text(f"@{tag}", style="magenta"))

            for task in items:
                if task["completed"]:
                    line = Text("  • ", style="dim")
                    line.append(f"{task['id']} ", style="dim green")
                    line.append("✔ ", style="green")
                    line.append(task["text"], style="dim")
                else:
                    line = Text("  • ", style="bold yellow")
                    line.append(f"{task['id']} ", style="bold yellow")
                    line.append(task["text"], style="bold white")

                console.print(line)

            console.print(Text(f"Completed: {completed_count} | Pending: {incomplete_count}", style="dim cyan"))

    # add task to json

    def addTaskJson(task: dict):
        data: dict = Functions.load_todos()

        if data['tasks']:
            new_id: str = str(max(map(int, data['tasks'].keys())) + 1)
        else:
            new_id: str = '1'

        data['tasks'][new_id] = task

        Functions.save_todos(data)
        print(f'\n[bold yellow]Task {new_id} Added![/bold yellow]\n')


    def build_task(task: str):
        task_data = {
            "task": task,
            "completed": False,
        }

        Functions.addTaskJson(task_data)

    def orderList():
        data = Functions.load_todos()
        tasks = data.get("tasks", {})

        if not tasks:
            return

        new_tasks = {}
        for i, (_, task_data) in enumerate(tasks.items(), start=1):
            new_tasks[str(i)] = task_data

        data["tasks"] = new_tasks
        Functions.save_todos(data)

        print("\n[bold green]Task list reordered![/bold green]\n")

    # mark task as done in json

    def doneTaskJson(doneIndex: list) -> str:

        data: dict = Functions.load_todos()

        try:
            if doneIndex[0] == "all":
                for key in data['tasks']:
                    data['tasks'][key]['completed'] = True

            else:
                for arg in doneIndex:
                
                    if "-" in arg:
                        min_i, max_i = arg.split("-")

                        for task in range(int(min_i), int(max_i) + 1):
                            task = str(task)
                            if task in data['tasks']:
                                data['tasks'][task]['completed'] = True

                    else:
                        data['tasks'][str(arg)]['completed'] = True

            Functions.save_todos(data)

            print(f'\n[bold yellow]Task(s) {doneIndex} marked Done![/bold yellow]\n')

        except ValueError:
            print('Invalid input. Please enter a valid number.')
        except KeyError:
            print('Invalid input. Please enter a valid number.')

    # remove tasks that are completed

    def clearTaskJson():

        data: dict = Functions.load_todos()

        for count in list(data['tasks']):
            if data['tasks'][count]['completed']:
                del data['tasks'][count]

        Functions.save_todos(data)

        print('\n[bold yellow]TODO list CLEARED![/bold yellow]\n')

    # print help commands

    def helpText() -> None:
        print(
            "[bold]Commands:[/bold]\n"
            "[cyan]  add[/cyan], a        Add new task\n"
            "[cyan]  done[/cyan], d       Mark task done\n"
            "[cyan]  list[/cyan], ls      Show todo list\n"
            "[cyan]  edit[/cyan], e       Edit task\n"
            "[cyan]  clear[/cyan], c      Clear done tasks\n"
            "[cyan]  order[/cyan], o      Order the Todos by IDs\n"
            "[cyan]  reload[/cyan], rld   Reload the app\n"
            "[cyan]  help[/cyan], h       Show help\n"
            "[cyan]  exit[/cyan], q       Exit app\n"
            "\n"
            "[bold]Batch operations:[/bold]\n"
            "[green]  done all[/green]    mark all tasks done\n"
            "[green]  done 2-4[/green]    mark tasks 2 3 4 done\n"
            "[green]  done 1 5 7[/green]  mark tasks 1, 5, 7 done\n"
            "[bold]Tags:[/bold]\n"
            "[green]  @tag[/green]    write the name of you tag after '@'\n"
        )

    # load json file

    def load_todos() -> dict:
        with open(todoJsonListPath(), 'r') as f:
            return json.load(f)

    # save to the json file

    def save_todos(data: dict):
        with open(todoJsonListPath(), 'w') as f:
            json.dump(data, f, indent=4)

    # load tags for autocompletion

    def load_tags() -> list[str]:
        data = Functions.load_todos()

        tags = set()
        pattern = re.compile(r'@[\w-]+')

        for task in data.get("tasks", {}).values():
            text = task.get("task", "")
            tags.update(pattern.findall(text))

        return sorted(tags)



from .shellcompleter import ShellCompleter
from .paths import todoHistoryFilePath

from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory

class Prompts:
        session = PromptSession(
        completer=ShellCompleter(),
        complete_while_typing=False,
        history=FileHistory(todoHistoryFilePath()),
    )



import sys
import requests
import subprocess
from importlib.metadata import version, PackageNotFoundError

PACKAGE_NAME = "todol"

class TodolUpgrade():
    def upgrade():
        try:
            installed_version = version(PACKAGE_NAME)
        except PackageNotFoundError:
            installed_version = "not installed"

        try:
            resp = requests.get(f"https://pypi.org/pypi/{PACKAGE_NAME}/json", timeout=3)
            resp.raise_for_status()
            latest_version = resp.json()["info"]["version"]
        except Exception:
            latest_version = "unknown"

        if installed_version == latest_version:
            print(f"{PACKAGE_NAME} {installed_version} Your packages are up to date")
        else:
            print(f"{PACKAGE_NAME} {installed_version} (newer version available: {latest_version})")

            update_question = input("Do you want to update it now? [Y/n]: ").strip().lower()
            if update_question in ("", "y", "yes"):
                try:
                    print("Running: pipx upgrade todol")
                    subprocess.run(["pipx", "upgrade", PACKAGE_NAME], check=True)
                    print("Update completed!")
                except FileNotFoundError:
                    print("pipx not found. Please install it first: https://pipx.pypa.io/latest/installation/")
                except subprocess.CalledProcessError:
                    print("Update failed. Try running the command manually: pipx upgrade todol")

        sys.exit(0)



from .commands import Commands

def aliases(func, *names):
    return {name: func for name in names}


COMMANDS = {
    "add": Commands.cmd_add,
    "done": Commands.cmd_done,
    "edit": Commands.cmd_edit,
    "help": Commands.cmd_help,
    "list": Commands.cmd_list,
    "clear": Commands.cmd_clear,
    "order": Commands.cmd_order,
    "reload": Commands.cmd_reload,
    "exit": Commands.cmd_exit,
}

ALIASES = {
    "a": "add",
    "d": "done",
    "e": "edit",
    "h": "help",
    "ls": "list",
    "c": "clear",
    "o": "order",
    "rld": "reload",
    "q": "exit",
}




from datetime import datetime
from platformdirs import user_data_dir
from pathlib import Path
from rich import print

# Todo files dir based on OS

DATA_DIR = Path(user_data_dir('todol', 'todol'))
TODO_DIR = DATA_DIR / 'todoFiles'

# Basic files for todol

TODO_JSON = TODO_DIR / 'main.json'
HISTORY_FILE = TODO_DIR / 'history'

# Backup dir

BACKUP_DIR = TODO_DIR / 'backups'

# Creating dictionaries

BACKUP_DIR.mkdir(parents=True, exist_ok=True)
TODO_DIR.mkdir(parents = True, exist_ok = True)

# Creating "blank" json

if not TODO_JSON.exists():
    TODO_JSON.write_text('{"tasks": {}}')

# Creating/Reseting history

HISTORY_FILE.touch()
HISTORY_FILE.write_text('')

def todoJsonListPath():
    return TODO_JSON

def todoHistoryFilePath():
    return HISTORY_FILE

def reset_todolist():
    print(
        '[bold red]!!! DANGER: THIS WILL ERASE ALL TODO DATA !!![/bold red]\n'
        'Make sure you have a backup first. You can run "todol --backup" or save a copy manually.\n'
    )

    choice = input('Do you really want to delete all tasks? [y/N] ').strip().lower()
    if choice in ('y', 'yes'):
        TODO_JSON.write_text('{"tasks": {}}')
        print('[bold green]All tasks have been cleared.[/bold green]')

    elif choice in ('', 'no', 'n'):
        print('[bold red]Operation cancelled. No data was lost.[/bold red]')


def backup_todolist():
    try:
        timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        default_backup = BACKUP_DIR / f'backup-{timestamp}.json'

        print(
            f'[bold yellow]Save backup somewhere else (Leave blank for default)[/bold yellow]\n'
            f'{default_backup}\n'
            )
        user_path = input('> ').strip()

        if user_path:
            backup_path = Path(user_path).expanduser()
            if backup_path.is_dir():
                backup_path = backup_path / default_backup.name
        else:
            backup_path = default_backup

        backup_path.parent.mkdir(parents=True, exist_ok=True)

        backup_path.write_text(TODO_JSON.read_text(), encoding='utf-8')

        print(f'[bold green]Todo list backed up to:[/bold green] {backup_path}')

    except KeyboardInterrupt:
        print('\n[bold red]No backup were made[/bold red]')

def load_backup():
    backups = sorted(BACKUP_DIR.glob('backup-*.json'))

    if not backups:
        print('[bold red]No backups found[/bold red]')
        return

    # List backups
    print('[bold yellow]Available backups:[/bold yellow]')
    for i, backup in enumerate(backups, 1):
        print(f'{i}. {backup.name}')

    # Select backup
    try:
        choice = int(input('\nSelect a backup number: '))
        selected_backup = backups[choice - 1]
    except (ValueError, IndexError):
        print('[bold red]Invalid selection[/bold red]')
        return
    except KeyboardInterrupt:
        print('\n[bold red]Load cancelled[/bold red]')
        return

    # Ask to backup current main.json
    print(
        '[bold yellow]This will overwrite main.json.[/bold yellow]\n'
        'Create a backup of the current todo list first?'
    )
    answer = input('(y/n) ').strip().lower()

    if answer == 'y':
        backup_todolist()

    try:
        TODO_JSON.write_text(selected_backup.read_text(), encoding='utf-8')
        print(f'[bold green]Backup loaded successfully:[/bold green] {selected_backup.name}')
    except KeyboardInterrupt:
        print('\n[bold red]Load cancelled[/bold red]')




import random
import asyncio
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.styles import Style
from prompt_toolkit.formatted_text import ANSI  # FIXED IMPORT
from prompt_toolkit.output.color_depth import ColorDepth
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from rich.align import Align
from io import StringIO

WIDTH = 30
HEIGHT = 15
TICK_RATE = 0.12

class SnakeGame:
    def __init__(self):
        self.reset()

    def reset(self):
        self.snake = [(WIDTH // 2, HEIGHT // 2)]
        self.direction = (1, 0)
        self.food = self.spawn_food()
        self.score = 0
        self.game_over = False

    def spawn_food(self):
        while True:
            pos = (random.randint(1, WIDTH - 2), random.randint(1, HEIGHT - 2))
            if pos not in self.snake:
                return pos

    def change_direction(self, dx, dy):
        if (dx, dy) != (-self.direction[0], -self.direction[1]):
            self.direction = (dx, dy)

    def update(self):
        if self.game_over:
            return

        head_x, head_y = self.snake[0]
        dx, dy = self.direction
        new_head = (head_x + dx, head_y + dy)

        if (
            new_head[0] <= 0
            or new_head[0] >= WIDTH - 1
            or new_head[1] <= 0
            or new_head[1] >= HEIGHT - 1
            or new_head in self.snake
        ):
            self.game_over = True
            return

        self.snake.insert(0, new_head)

        if new_head == self.food:
            self.score += 1
            self.food = self.spawn_food()
        else:
            self.snake.pop()

    def render(self):
        grid = [[" " for _ in range(WIDTH)] for _ in range(HEIGHT)]

        for x in range(WIDTH):
            grid[0][x] = "#"
            grid[HEIGHT - 1][x] = "#"
        for y in range(HEIGHT):
            grid[y][0] = "#"
            grid[y][WIDTH - 1] = "#"

        fx, fy = self.food
        grid[fy][fx] = "●"

        for i, (x, y) in enumerate(self.snake):
            grid[y][x] = "█" if i == 0 else "▓"

        lines = ["".join(row) for row in grid]
        board = "\n".join(lines)

        text = Text()
        text.append(board, style="bold green")
        text.append(f"\n\nScore: {self.score}", style="bold yellow")

        if self.game_over:
            text.append(
                "\n\nGAME OVER - Press R to Restart or Q to Quit",
                style="bold red",
            )

        return Align.center(
            Panel(text, title="Snake", border_style="bright_blue")
        )

async def run_snake():
    game = SnakeGame()
    kb = KeyBindings()

    @kb.add("up")
    def _(event):
        game.change_direction(0, -1)

    @kb.add("down")
    def _(event):
        game.change_direction(0, 1)

    @kb.add("left")
    def _(event):
        game.change_direction(-1, 0)

    @kb.add("right")
    def _(event):
        game.change_direction(1, 0)

    @kb.add("r")
    def _(event):
        if game.game_over:
            game.reset()

    @kb.add("q")
    def _(event):
        event.app.exit()

    def get_text():
        buffer = StringIO()
        temp_console = Console(
            file=buffer,
            force_terminal=True,
            color_system="standard",
        )
        temp_console.print(game.render())
        return ANSI(buffer.getvalue())

    app = Application(
        layout=Layout(Window(FormattedTextControl(get_text))),
        key_bindings=kb,
        full_screen=True,
        style=Style.from_dict({
            "": "bg:#000000 #ffffff",
        }),
        color_depth=ColorDepth.DEPTH_8_BIT,
    )

    async def game_loop():
        while True:
            await asyncio.sleep(TICK_RATE)
            game.update()
            app.invalidate()

    asyncio.create_task(game_loop())
    await app.run_async()

def main():
    asyncio.run(run_snake())




from platformdirs import user_data_dir
from pathlib import Path

class TodolPath():
    def path():
        DATA_DIR = Path(user_data_dir('todol', 'todol'))
        TODO_DIR = DATA_DIR / 'todoFiles'

        if TODO_DIR.exists():
            print(f"{TODO_DIR}")
        else:
            print("No data directory found. Run the app first: todol")



from prompt_toolkit.completion import Completer, Completion

from prompt_toolkit.completion import Completer, Completion

class ShellCompleter(Completer):
    def get_completions(self, document, complete_event):

        from .commands_list import COMMANDS
        from .functions import Functions
        # tags.py
        TAGS = Functions.load_tags()

        if not complete_event.completion_requested:
            return

        text = document.text_before_cursor
        words = text.split()

        if not words:
            for cmd in COMMANDS:
                yield Completion(cmd, start_position=0)
            return

        last = words[-1]

        if last.startswith("@"):
            current = last
            for tag in TAGS:
                if tag.startswith(current):
                    yield Completion(tag, start_position=-len(current))
            return

        if len(words) == 1 and not text.endswith(" "):
            current = words[0]
            for cmd in COMMANDS:
                if cmd.startswith(current):
                    yield Completion(cmd, start_position=-len(current))
            return
        return
